//=====================================================================================
// Program: Temperatur_Feuchte Logger mit SD Karte
//
// Author : Daniel Buergin
//
// Beschreibung: Auswertung von SHT75 Sensoren und Anzeige auf 7-Seg. LED's
//               Speicherung der Daten auf SD Karte
//
//=====================================================================================
// Version: 0.9
//=====================================================================================
// Aenderungen:
//
// 24-Jan-2009 bue Erste funktionierende Version
//=====================================================================================
//
// Copyrights: - UART und LCD Library von Peter Fleury (http://jump.to/fleury)
//             - SD/MMC von Roland Riegel (www.roland-riegel.de)
//             - SHT75 Brian Low (mportobello@hotmail.com) / Daniel Buergin
//             - MAX7219 Code Daniel Buergin
//             - Debounce Routinen von Peter Dannegger
//
//=====================================================================================
//
// HW: - ATmega128 Board von www.futurlec.com
//     - 4 x Kingsbirght 7-Segment LED an MAX7219
//     - 1 x Sensirion SHT75 Temperatur/Feuchte Sensor
//     - SD/MMC Kartenmodul fuer 5V von www.avrcard.com
//
//=====================================================================================
//
// Konfiguration:
// ==============
//
// - PortA:
//      PA0: ()
//      PA1: ()
//      PA2: ()
//      PA3: ()
//      PA4: () Button1 (mit 10kohm externem Pullup)
//      PA5: () MAX7219 IN
//      PA6: () MAX7219 LOAD
//      PA7: () MAX7219 CLK 
//
// - PortB:
//      PB0: (SS)   SS  (Pin 2 auf AVRCard SD/MMC Modul)
//      PB1: (SCK)  SCK (Pin 3 auf AVRCard SD/MMC Modul)
//      PB2: (MOSI) SO  (Pin 5 auf AVRCard SD/MMC Modul)
//      PB3: (MISO) SI  (Pin 4 auf AVRCard SD/MMC Modul)
//      PB4: ()
//      PB5: ()
//      PB6: ()
//      PB7: ()
//
// - PortC: 
//      PC0: ()
//      PC1: ()
//      PC2: ()
//      PC3: ()
//      PC4: ()
//      PC5: ()
//      PC6: () SHT75 Data
//      PC7: () SHT75 Clock
//
// - PortD:
//      PD0: ()
//      PD1: ()
//      PD2: (UART1,RX)
//      PD3: (UART1,TX)
//      PD4: ()
//      PD5: ()
//      PD6: ()
//      PD7: ()
//
// - PortE:
//      PE0: (UART0,RX)
//      PE1: (UART1,TX)
//      PE2: ()
//      PE3: ()
//      PE4: (INT4)
//      PE5: (INT5)
//      PE6: (INT6)
//      PE7: (INT7)
//
// - PortF:
//      PF0: (ADC0)
//      PF1: (ADC1)
//      PF2: (ADC2)
//      PF3: (ADC3)
//      PF4: (ADC4)
//      PF5: (ADC5)
//      PF6: (ADC6)
//      PF7: (ADC7)
//=====================================================================================

// Includes 

#include "main.h" 
#include "uart.h" 
#include "sht75.h" 
#include "max7219.h" 
#include "fat.h"
#include "fat_config.h"
#include "partition.h"
#include "sd_raw.h"
#include "sd_raw_config.h"
#include "file.h"

struct fat_file_struct* fd;
struct partition_struct* partition;
struct fat_fs_struct* fs;
struct fat_dir_entry_struct directory;
struct fat_dir_struct* dd;
struct fat_dir_entry_struct dir_entry;
struct fat_dir_entry_struct file_entry;

//=====================================================================================
// Main Routine
//=====================================================================================

int main(void)
{

	sht75_array sht75;
	uint8_t sht_ok = 0;

	char temp[7];
	char humi[7];
	uint8_t sht75_status;
	uint8_t sht75_error;
	uint8_t disp_state = 0;

	uint8_t  sd_counter = 0;
	char    sd_buffer[80];
	uint8_t  sd_log = 0;

	uint32_t local_timer = 0;

	// Setup the Button	
	setup_buttons();

	// Start the Timers
	init_timer0();

	// Start the Temp/Hum Sensor
	ShtInit();

	// Initialize the MAX7219 Chip
	max7219_Init();

	// Init UART0 with 9600 for Debugging Output
	uart_init( UART_BAUD_SELECT(UART0_BAUD, F_CPU));

	// Die Interrupts einschalten
	sei();

	my_delay_10ms(10); // 100ms warten, damit der SHT75 einschwingen kann

	uart_puts_P("Temperatur / Feuchte Logger Version: 0.9 by D.Buergin\n\r");

	// Version auch auf den 7-Seg. anzeigen

	max7219_Write(1,'U');
	max7219_Write(2,' ');
	max7219_WriteDP(3,'0');
	max7219_Write(4,'9');

	my_delay_10ms(100);

	// Hauptschleife

	while(1)
	{
		// Wenn Button gedrueckt, den "sd_log" Switch toggeln
		// 0 = Nicht loggen, resp. File schliessen wenn vorher gelogged wurde
		// 1 = File oeffnen und alle 60s einen Datensatz loggen
		
		if (get_key_press(1<<B1))
		{
			if (sd_log == 0)
			{
				max7219_Write(1,'O');
				max7219_Write(2,'n');
				max7219_Write(3,' ');
				max7219_Write(4,' ');
				sd_log = 1;
				open_logging();
			}
			else
			{
				max7219_Write(1,'O');
				max7219_Write(2,'F');
				max7219_Write(3,'F');
				max7219_Write(4,' ');
				sd_log = 0;
				close_logging();
			}
			my_delay_10ms(20);
		}

		// Nun wird in einem 3 Sekunden Takt gearbeitet (disp_state)
		// 1s = Messen
		// 2s = Feuchte anzeigen
		// 3s = Temperatur anzeigen
		//
		// Dadurch erscheint die Temperatur doppelt so lange auf dem
		// Display, da beim erneuten messen noch die alte Temp auf
		// dem Display steht. 
		// Ist gewollt so ;-) da mich die Temp. mehr interessiert
		
		if (local_timer != sec_timer)
		{
			local_timer = sec_timer;
			sd_counter++;

			// Messen
			
			if (disp_state == 0)
			{
				disp_state = 1;
				sht_ok = 0;

				// Den SHT75 reseten. Braucht ev. nicht jedes Mal gemacht werden. 
		
				ShtReset();
	
				// Die eigentliche Messung machen

				ShtReadEverything(&sht75);
				sht75_error = ShtReadStatus(&sht75_status);

				// Wenn kein Fehler beim messen aufgetaucht ist

				if (sht75_error == 0)
				{
					sht_ok = 1;

					// Beide Werte in einen String, so dass wir sie Zeichen fuer Zeichen anzeigen koennen
					sprintf(temp,"%+02.1f",sht75.Temperature);
					sprintf(humi,"%02.1f",sht75.Humidity);

					// Alles auf UART0 ausgeben
					sprintf(disp_buffer,"Temp: %+02.2f Feuchte: %2.2f Taupunkt: %+02.2f Status: %02x Error: %02x\n\r",
						sht75.Temperature,
						sht75.Humidity,
						sht75.Dewpoint,
						sht75_status,
						sht75_error);

					uart_puts(disp_buffer);

					// Wenn eine Minute um ist, alles auf der SD speichern
		
					if (sd_counter > 59)
					{
						sd_counter = 0;

						// Wenn das Logging eingeschaltet ist

						if (sd_log == 1)
						{
							sprintf(sd_buffer,"%+02.2f;%2.2f;%+02.2f\n",
								sht75.Temperature,
								sht75.Humidity,
								sht75.Dewpoint);

							if(fat_write_file(fd, (uint8_t*) sd_buffer, strlen(sd_buffer)) != strlen(sd_buffer))
							{
								uart_puts_P("Fehler beim schreiben\n\r");
							}
							else
							{
								uart_puts_P("Datensatz erfolgreich geschrieben\n\r");
							}
							sd_raw_sync();
						}
					}
				}
				else
				{
					uart_puts_P("Fehler beim SHT75 Sensor !!!!!\n\r");
				}
			}
			else if (disp_state == 1)
			{
				// Feuchte anzeigen
				
				disp_state = 2;
		
				max7219_Write(1,' ');
				max7219_Write(2,humi[0]);
				max7219_WriteDP(3,humi[1]);
				max7219_Write(4,humi[3]);
			}
			else
			{
				// Temperatur anzeigen

				disp_state = 0;

				// Schauen, ob die Temp. im Minusbereich ist, 
				// wenn ja, auf der ersten 7-Seg. ein Minus anzeigen
				
				if (temp[0] == '-')
				{
					max7219_Write(1,'-');
				}
				else
				{
					max7219_Write(1,' ');
				}

				max7219_Write(2,temp[1]);
				max7219_WriteDP(3,temp[2]);
				max7219_Write(4,temp[4]);
			}
		}
	}
}

//=====================================================================================
// Timer0 und Timer1 initialisieren
// - Timer0 laeuft im Millisekundentakt fuer die Tastenabfrage
// - Timer1 laeuft im Sekundentakt fuer Anzeige
//=====================================================================================

void init_timer0(void)
{
	// ====================================================
	// Timer 0 laueft im Millisekunden Takt
	// ====================================================

    TCCR0 = 0;

    // Timer/Counter Interrupt Flag Register (Bit 0 und 1 werden gesetzt)
 
   	TIFR |= _BV(OCF0) | _BV(TOV0);

    // Output Compare Interrupt enablen

	TIMSK |= _BV(TOIE0)|_BV(OCIE0);

    // CTC Bit im TCCR0 Register, CS00 und CS02 Bit setzen
    // d.h. prescale = 128 oder 10us Tick bei 16Mhz
    // OCR0 wird auf 125 gesetzt. 10us mal 125 = 1ms

    TCCR0  = _BV(WGM01)|_BV(CS02)|_BV(CS00);
    TCNT0  = 0;
    OCR0   = OCR_1MS;   // 125

	// ====================================================
	// Timer 1 laueft im Sekunden Takt
	// ====================================================

	TCCR1B = 0;

	// 14.5 Mhz: Prescale auf 1024, OCR1A auf 14400 ergibt 1 Sekunde
	// 16.0 Mhz: Prescale auf 1024, OCR1A auf 15625 ergibt 1 Sekunde
   
	TCCR1B = _BV(WGM12) | _BV(CS12) | _BV(CS10);

	OCR1A = 15625;

	TIMSK |= _BV(TOIE1) | _BV(OCIE1A);
}

//=====================================================================================
// Initialisieren der Taster
//=====================================================================================

void setup_buttons(void)
{
    ButtonDDR  &= ( 1 << B1 );
    ButtonPort |= ( 1 << B1 );
}

//=====================================================================================
// Timer0 Routine
// Routine stammt von Peter Danneger und enthaelt die
// Entprellroutinen fuer die Taster
//=====================================================================================

ISR (TIMER0_COMP_vect)
{
    static uint8_t ct0, ct1, rpt;
    uint8_t i;

    i = key_state ^ ~ButtonPIN;     // key changed ?
    ct0 = ~( ct0 & i );         // reset or count ct0
    ct1 = ct0 ^ (ct1 & i);      // reset or count ct1
    i &= ct0 & ct1;         // count until roll over ?
    key_state ^= i;         // then toggle debounced state
    key_press |= key_state & i;     // 0->1: key press detect

    if( (key_state & REPEAT_MASK) == 0 )    // check repeat function
    {
        rpt = REPEAT_START;     // start delay
    }

    if( --rpt == 0 )
    {
        rpt = REPEAT_NEXT;          // repeat delay
        key_rpt |= key_state & REPEAT_MASK;
    }
}

//=====================================================================================
// Timer1 Routine
// Dieser Timer laeuft im Sekundentakt.
//=====================================================================================

ISR(TIMER1_COMPA_vect)
{
	// Counter um im Hauptprogramm rauszufinden, wenn eine Sekunde verstrichen
	// ist.
	sec_timer++;
}

//=====================================================================================
// debounce()
//  - Taster abfragen
//=====================================================================================


uint8_t get_key_press( uint8_t key_mask )
{
	uint8_t sreg;

	// Register sichern
	sreg = SREG;
	// Interrupts sperren
	cli(); 
	  key_mask &= key_press;  // read key(s)
	  key_press ^= key_mask;  // clear key(s)

	// Interrupts wieder freigeben, sei() braucht es nicht
	SREG = sreg;

	return key_mask;
}

//=====================================================================================
// Fuer etwas groessere Delay
//=====================================================================================

void my_delay_10ms(uint8_t t)
{
	if (t==0) return;
	while(t--)
	{
		_delay_ms(10);
	}
}
	
//=====================================================================================
// Logging File auf SD-Karte schliessen. 
//=====================================================================================

void close_logging(void)
{
	// Zur Sicherheit noch mal einen sync, damit ist das FS konsistent
	sd_raw_sync();
	fat_close_file(fd);
	fat_close(fs);
	partition_close(partition);
}

//=====================================================================================
// Logging File auf SD-Karte oeffnen und auf das Ende positionieren. 
//=====================================================================================

void open_logging(void)
{
	// Das Logging File auf der SD Karte oeffnen
	
	if(!sd_raw_init())
	{
		uart_puts_P("Fehler bei Initialisierung von MMC/SD Karte\n\r");
	}
	else
	{
		// open first partition

		partition = partition_open(sd_raw_read,
			sd_raw_read_interval,
			sd_raw_write,
			sd_raw_write_interval,
			0
		);

		if(!partition)
		{
			// If the partition did not open, assume the storage device
			// is a "superfloppy", i.e. has no MBR.

			partition = partition_open(sd_raw_read,
				sd_raw_read_interval,
				sd_raw_write,
				sd_raw_write_interval,
				-1
			);

			if(!partition)
			{
				uart_puts_P("Kann Partition nicht oeffnen\n\r");
			}
		}

		// open file system

		fs = fat_open(partition);
		if(!fs)
		{
			uart_puts_P("Kann FAT nicht oeffnen\n\r");
		}

		// open root directory

		fat_get_dir_entry_of_path(fs, "/", &directory);

		dd = fat_open_dir(fs, &directory);
		if(!dd)
		{
			uart_puts_P("Kann Root-Verzeichnis nicht oeffnen\n\r");
		}

		// print some card information as a boot message
		print_disk_info(fs);

		// print directory entry
		uart_puts_P("Directory:\n\r");
		while(fat_read_dir(dd, &dir_entry))
		{
			uint8_t spaces = sizeof(dir_entry.long_name) - strlen(dir_entry.long_name) + 4;

			uart_puts(dir_entry.long_name);
			uart_putc(dir_entry.attributes & FAT_ATTRIB_DIR ? '/' : ' ');
			while(spaces--)
				uart_putc(' ');
			uart_putdw_dec(dir_entry.file_size);
			uart_puts("\n\r");
		}

		// Logger File oeffnen und an das Ende spulen, so dass etwas angehaengt werden kann
		// Wenn das File noch nicht besteht, es erstellen

		uart_puts_P("Oeffne File \"logger.csv\":\n\r");
		fd = open_file_in_dir(fs, dd, "logger.csv");
		if(!fd)
		{
			uart_puts_P("Fehler beim oeffnen von: \"logger.csv\"\n\r");

			// Nun ein neues File erstellen

			if(!fat_create_file(dd, "logger.csv", &file_entry))
			{
				uart_puts_P("Fehler beim anlegen von: \"logger.csv\"\n\r");
			}
			else
			{
				uart_puts_P("File: \"logger.csv\" erfolgreich angelegt\n\r");
				sd_raw_sync();
			}

			fd = open_file_in_dir(fs, dd, "logger.csv");
			if(!fd)
			{
				uart_puts_P("2. Fehler beim oeffnen von: \"logger.csv\"\n\r");
			}
		}

		if(fd)
		{
			// Auf das Ende des Files positionieren
			// FAT_SEEK_END und offset=0 heissen, vom Ende der Datei aus 0 Bytes lesen,
			// mit anderen Worten, an das Ende des Files gehen.
			
			int32_t write_offset = 0;

			if(!fat_seek_file(fd, &write_offset, FAT_SEEK_END))
			{
				uart_puts_P("Fehler bei \"fat_seek_file\": \"logger.csv\"\n\r");
				fat_close_file(fd);
			}
			else
			{
				// Header schreiben
				sprintf(sd_header,"Temperatur (Celsius); Feuchte (%%); Taupunkt (Celsius)\n");
				if(fat_write_file(fd, (uint8_t*) sd_header, strlen(sd_header)) != strlen(sd_header))
				{
					uart_puts_P("Fehler beim schreiben\n\r");
				}
				sd_raw_sync();
			}
		}
	}
}
